package asyncLog

import (
	"github.com/astaxie/beego"
	"github.com/astaxie/beego/logs"
	"github.com/ethereum/go-ethereum/accounts/abi"
	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/core/types"
	"github.com/ethereum/go-ethereum/ethclient"
	"hashrateNode/cache"
	hashrateCommon "hashrateNode/common"
	"hashrateNode/commonApi"
	vm "hashrateNode/contract/VmContract"
	"hashrateNode/log"
	"hashrateNode/models"
	"hashrateNode/utils"
	"math/big"
	"strings"
	"time"
)

func VmStatusContractHandler(vLog types.Log) error {
	logs.Info("handle vm create contract logs.")
	vmAbiInfo, _ := abi.JSON(strings.NewReader(vm.VmABI))
	{
		method := vLog.Topics[0]
		switch method.String() {
		case hashrateCommon.VmStatusEvent:
			{
				vmStatusEvent := &models.VmStatusEvent{}
				err = vmAbiInfo.UnpackIntoInterface(vmStatusEvent, "UpdateVmStatus", vLog.Data)
				if err != nil {
					err := cache.Redis.LpushByte(hashrateCommon.EventList, utils.Marshal(vmStatusEvent))
					if err != nil {
						log.Error("cache redis lPush Byte error:", err.Error())
						return err
					}
				}
				isSuccess, err := handlerVm(vmStatusEvent)
				if err != nil {
					return err
				}
				if isSuccess {
					log.Info("VM status update successful")
					return nil
				}
			}
		}
	}
	return nil
}

func handlerVm(vmStatus *models.VmStatusEvent) (bool, error) {
	ticket, err := commonApi.GetCreateVmTicket()
	if err != nil {
		log.Error("GetCreateVmTicket error:", err)
		return false, err

	}
	vmInfo := &models.VM{
		NewId:      vmStatus.VmId.String(),
		HeaderInfo: ticket,
	}
	switch vmStatus.Status {
	case hashrateCommon.VmStatusEnabled:
		{
			_, err := commonApi.StartVm(vmInfo)
			if err != nil {
				log.Error("Start vm error:", err)
				return false, err
			}
		}
	case hashrateCommon.VmStatusDisable:
		{
			_, err := commonApi.ShutdownVm(vmInfo)
			if err != nil {
				log.Error(" Shutdown vm error:", err)
				return false, err
			}
		}
	case hashrateCommon.VmStatusReboot:
		{
			_, err := commonApi.RebootVm(vmInfo)
			if err != nil {
				log.Error("Reboot vm error:", err)
				return false, err
			}
		}
	case hashrateCommon.VmStatusDelete:
		{

		}

	}
	vmCurrentStatus, err := commonApi.VmStatus(vmInfo)
	if err != nil {
		log.Error("VmStatus error:", err)
		return false, err
	}

	rpcUrl := beego.AppConfig.String("chain_rpc_url")
	client, err := ethclient.Dial(rpcUrl)
	if err != nil {
		log.Error("Eth client.Dial error:", err)
		return false, err
	}
	defer client.Close()
	nonce, err = utils.EthGetTransactionCount(client, fromAddr)
	if err != nil {
		log.Error("EthGetTransactionCount error:", err)
		return false, err
	}
	if (vmStatus.Status == hashrateCommon.VmStatusDisable) && vmCurrentStatus.Status != hashrateCommon.Stopped {
		time.Sleep(time.Second * 10)
	}
	if (vmStatus.Status == hashrateCommon.VmStatusEnabled || vmStatus.Status == hashrateCommon.VmStatusReboot) && vmCurrentStatus.Status != hashrateCommon.Running {
		time.Sleep(time.Second * 10)
	}
	if vmStatus.Status == hashrateCommon.VmStatusReboot {
		vmStatus.Status = hashrateCommon.VmStatusEnabled
	}
	vmContract, err := vm.NewVm(common.HexToAddress(hashrateCommon.VmContract), client)
	if err != nil {
		log.Error("NewVm error:", err)
		return false, err
	}
	nonceBigInt := new(big.Int).SetUint64(nonce)
	tranOpts.Nonce = nonceBigInt
	status, err := vmContract.UpdateVmStatus(tranOpts, vmStatus.VmId, vmStatus.Status)
	if err != nil {
		log.Error("UpdateVmStatus error:", err)
		return false, err
	}
	log.Info("UpdateVmStatus success:", status.Hash())
	return true, nil
}
